from dataclasses import dataclass
from typing import Dict, List, Union

import dash_bootstrap_components as dbc
from dash import dcc, html

from .base_component import BaseComponent


@dataclass
class CustomDropdown(BaseComponent):
    id: str
    size: str
    label: str
    options: List[Union[Dict[str, str], Dict[str, int], Dict[int, int], Dict[int, str]]]
    default_value: Union[int, str, List[int], List[str]]
    multi: bool = False
    searchable: bool = False
    clearable: bool = False

    @property
    def layout(self):
        return html.Div(
            className=f"custom-dropdown-style {self.size}",
            children=[
                dbc.Label(self.label, html_for=self.id),
                dcc.Dropdown(
                    id=self.id,
                    options=self.options,
                    multi=self.multi,
                    value=self.default_value,
                    searchable=self.searchable,
                    clearable=self.clearable,
                ),
            ],
        )



# default name of the settings file which defines an DasboardApp application
DEFAULT_SETTINGS_FILE = "settings.py"

# default directory for implementation of dashboards
DEFAULT_DASHBOARDS_DIR = "dashboards"

# default name of the main file which runs the app
DEFAULT_RUNSERVER_FILE = "runserver.py"

# default name of the assets folder
DEFAULT_ASSETS_DIR = "assets"



from pathlib import Path


def valid_dir_name(name: str) -> str:
    name = name.lower()
    if not name.isidentifier():
        raise ValueError(
            "Please enter valid name containing only characters and underscores."
        )
    return name


def validate_path_does_not_exist(p: Path) -> bool:
    if p.exists():
        raise FileExistsError(f"Directory with path '{p}' already exists.")
    return True



from pathlib import Path

import click

from .assets import generate_assets
from .config import (
    DEFAULT_ASSETS_DIR,
    DEFAULT_DASHBOARDS_DIR,
    DEFAULT_RUNSERVER_FILE,
    DEFAULT_SETTINGS_FILE,
)
from .utils import valid_dir_name, validate_path_does_not_exist

SETTINGS_CONTENT = """import pathlib

from dash_dashboards.app import DashboardApp, MenuGroup, MenuItem

ASSETS_DIR = (pathlib.Path(__file__).parent / "assets").resolve()

app = DashboardApp(
    title="{project_name}",
    menu=[],
    assets_folder=ASSETS_DIR,
)
"""

RUNSERVER_CONTENT = """from dash_dashboards.commands import run_server

from settings import app

run_server(obj=app)
"""

WSGI_CONTENT = """# !pip install -r requirements.txt

import os

from settings import app

app.run_server(port=int(os.environ["CDSW_READONLY_PORT"]))
"""


@click.command(name="startproject")
@click.option("-n", "--name", prompt="Project name", type=str, help="Project name.")
@click.option(
    "-p",
    "--dir-path",
    prompt="Directory path",
    type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=Path),
    default=Path("./").resolve(),
    help="Path to the directory where project will be created.",
)
def start_project(name: str, dir_path: Path):
    project_dir_name = valid_dir_name(name)
    project_path = dir_path / project_dir_name

    validate_path_does_not_exist(project_path)
    project_path.mkdir()

    settings_path = project_path / DEFAULT_SETTINGS_FILE
    runserver_file = project_path / DEFAULT_RUNSERVER_FILE
    dashboards_path = project_path / DEFAULT_DASHBOARDS_DIR
    wsgi_file = project_path / "wsgi.py"
    assets_path = project_path / DEFAULT_ASSETS_DIR

    dashboards_path.mkdir()
    with settings_path.open("w") as fp:
        fp.writelines(SETTINGS_CONTENT.format(project_name=name))

    with runserver_file.open("w") as fp:
        fp.writelines(RUNSERVER_CONTENT)

    with wsgi_file.open("w") as fp:
        fp.writelines(WSGI_CONTENT)

    generate_assets(assets_path)



from abc import ABC, abstractmethod

from dash.development.base_component import Component


class BaseComponent(ABC):
    @property
    @abstractmethod
    def layout(self) -> Component:
        pass



import click

from .commands import generate_assets_cmd, start_dash, start_project


@click.group()
def cli():
    pass


cli.add_command(start_project)
cli.add_command(start_dash)
cli.add_command(generate_assets_cmd)

if __name__ == "__main__":
    cli()  # pragma: no cover



from dataclasses import dataclass

from dash import dcc, html

from .base_component import BaseComponent


@dataclass
class NotFound(BaseComponent):
    id: str
    size: str = "col-6"

    @property
    def layout(self):
        return html.Div(
            className="page-container",
            children=[
                html.Div(
                    className="row g-4 page-content d-flex justify-content-center",
                    children=[
                        html.Div(
                            id=self.id,
                            className=f"""content-container not-found-content
                            align-center {self.size}""",
                            children=[
                                html.Img(
                                    className="not-found-img",
                                    src="../assets/img/wrong-way.svg",
                                    height=100,
                                ),
                                html.Div(
                                    className="not-found-text",
                                    children=[
                                        "Oops! You've followed the wrong route! :("
                                    ],
                                ),
                                html.Div(
                                    children=[
                                        dcc.Link(
                                            className="""list-group-item border-0
                                            d-inline-block text-truncate nav-item
                                            not-found-link auto-width""",
                                            href="/",
                                            children=[
                                                html.I(className="bi bi-arrow-left"),
                                                html.Span(
                                                    "Get back to the right track."
                                                ),
                                            ],
                                        ),
                                    ]
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )



from dataclasses import dataclass

from dash import html

from .base_component import BaseComponent


@dataclass
class Kpi(BaseComponent):
    id: str
    size: str
    kpi_value: int
    kpi_text: str
    kpi_unit: str = None
    double_kpi: bool = False
    secondary_id: str = None
    secondary_kpi_value: int = None
    secondary_kpi_text: str = None
    secondary_kpi_unit: str = None

    @property
    def layout(self):
        if self.double_kpi:
            return self._get_double_kpi_layout()
        return self._get_single_kpi_layout()

    def _get_single_kpi_layout(self):

        return html.Div(
            className=self.size,
            children=[
                html.Div(
                    className="content-container kpi-container",
                    children=[
                        html.Div(
                            className="row text-center",
                            children=[
                                html.Div(
                                    className="col kpi-body",
                                    children=[
                                        html.Span(
                                            id=self.id, children=[self.kpi_value]
                                        ),
                                        html.Span(self.kpi_unit),
                                    ],
                                ),
                            ],
                        ),
                        html.Div(
                            className="row container-title",
                            children=[
                                html.Div(
                                    className="col kpi-title text-center",
                                    children=[self.kpi_text],
                                )
                            ],
                        ),
                    ],
                )
            ],
        )

    def _get_double_kpi_layout(self):
        return html.Div(
            className=self.size,
            children=[
                html.Div(
                    className="content-container kpi-container",
                    children=[
                        html.Div(
                            className="row",
                            children=[
                                html.Div(
                                    className="col main-kpi text-end kpi-body",
                                    children=[
                                        html.Span(
                                            id=self.id, children=[self.kpi_value]
                                        ),
                                        html.Span(self.kpi_unit),
                                    ],
                                ),
                                html.Div(
                                    className="col secondary-kpi",
                                    children=[
                                        html.Div(
                                            className="row",
                                            children=[
                                                html.Div(
                                                    className="col secondary-kpi-body",
                                                    children=[
                                                        html.Span(
                                                            id=self.secondary_id,
                                                            children=[
                                                                self.secondary_kpi_value
                                                            ],
                                                        ),
                                                        html.Span(
                                                            self.secondary_kpi_unit
                                                        ),
                                                    ],
                                                ),
                                                html.Div(
                                                    className="row secondary-kpi-title",
                                                    children=[self.secondary_kpi_text],
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        html.Div(
                            className="row container-title",
                            children=[
                                html.Div(
                                    className="col kpi-title text-center",
                                    children=[self.kpi_text],
                                ),
                            ],
                        ),
                    ],
                )
            ],
        )



from typing import Callable, List

import dash_bootstrap_components as dbc
from dash import Dash, Input, Output, html
from dash.development.base_component import Component

from ..components.not_found import NotFound
from .base_layout import Menu, MenuItem, get_main_layout

BootstrapTheme = str
Icons = "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.7.0/font/bootstrap-icons.css"


class DashboardApp(Dash):
    def __init__(
        self,
        menu: Menu = [],
        name=None,
        theme: BootstrapTheme = dbc.themes.SANDSTONE,
        server=True,
        assets_folder="assets",
        assets_url_path="assets",
        assets_ignore="",
        assets_external_path=None,
        eager_loading=False,
        include_assets_files=True,
        url_base_pathname=None,
        requests_pathname_prefix=None,
        routes_pathname_prefix=None,
        serve_locally=True,
        compress=None,
        meta_tags=None,
        external_scripts=None,
        external_stylesheets=None,
        suppress_callback_exceptions=True,
        prevent_initial_callbacks=False,
        show_undo_redo=False,
        extra_hot_reload_paths=None,
        plugins=None,
        title="Dash",
        update_title="Updating...",
        long_callback_manager=None,
        **obsolete,
    ):
        self._menu = menu

        default_style = [
            theme,
            Icons,
            "assets/style.css",
        ]

        if not external_stylesheets:
            external_stylesheets = default_style
        else:
            external_stylesheets = default_style + external_stylesheets

        super().__init__(
            name=name,
            server=server,
            assets_folder=assets_folder,
            assets_url_path=assets_url_path,
            assets_ignore=assets_ignore,
            assets_external_path=assets_external_path,
            eager_loading=eager_loading,
            include_assets_files=include_assets_files,
            url_base_pathname=url_base_pathname,
            requests_pathname_prefix=requests_pathname_prefix,
            routes_pathname_prefix=routes_pathname_prefix,
            serve_locally=serve_locally,
            compress=compress,
            meta_tags=meta_tags,
            external_scripts=external_scripts,
            external_stylesheets=external_stylesheets,
            suppress_callback_exceptions=suppress_callback_exceptions,
            prevent_initial_callbacks=prevent_initial_callbacks,
            show_undo_redo=show_undo_redo,
            extra_hot_reload_paths=extra_hot_reload_paths,
            plugins=plugins,
            title=title,
            update_title=update_title,
            long_callback_manager=long_callback_manager,
            **obsolete,
        )

        self.layout = self.create_layout()
        self.register_callbacks()

    @property
    def pages(self) -> List[MenuItem]:
        return [page for item in self._menu for page in item.pages]

    def create_navigation_layout(self) -> List[Component]:
        app_title = html.Div(
            self.title,
            className="mx-3 py-4 border-bottom app-title",
        )
        navigation_items = [item.get_menu_item_layout() for item in self._menu]
        navigation = html.Div(
            className="p-3 sidebar-menu",
            children=[
                html.Ul(navigation_items, className="nav flex-column"),
            ],
        )
        return [app_title, navigation]

    def create_layout(self) -> Component:
        return get_main_layout(self.create_navigation_layout())

    def register_callbacks(self):
        """
        Implements generic callbacks common for all apps,
        here used for navigation

        Example:
        ```
            @self.callback(
                Output("nav", "children"), Input("nav-item", "value")
            )
            def common_nav_callback(nav_item_value):
                ...
                return nav_children
        ```
        """

        @self.callback(
            Output("content", "children"),
            [Input("url", "pathname")],
        )
        def display_page(pathname):
            path = pathname.rstrip("/")

            for page in self.pages:
                if page.route.rstrip("/") == path:
                    return page.get_layout()

            return NotFound("not-found").layout

    def add_callbacks(self, callbacks: Callable[[Dash], None]):
        callbacks(self)



from pathlib import Path

import click

from dash_dashboards.commands.utils import valid_dir_name, validate_path_does_not_exist

from .config import DEFAULT_DASHBOARDS_DIR

BASE_APP_CONTENT = """from dash import Dash, Input, Output, html


layout = html.Div(
    className="page-container",
    children=[
        html.Div(
            className="row page-header gy-6",
            children=[
                html.Div(
                    className="col-3 page-title",
                    children=[html.Span("< alarm_page title section >")],
                ),
                html.Div(
                    className="col-9", children=["< PLACEHOLDER FOR PAGE FILTERS >"]
                ),
            ],
        ),
        html.Div(
            className="row g-4 page-content",
            children=[
                html.Div(
                    className="content-container col-12",
                    children=["< PLACEHOLDER FOR PAGE CONTENT >"],
                )
            ],
        ),
    ],
)

def callbacks(app: Dash):
    @app.callback(
        Output("output-id", "output_attribute"),
        [Input("input-id", "input_attribute")],
    )
    def callback_handler(input_attribute):
        output_attribute = html.H2("Callback output")
        return output_attribute
"""


@click.command(name="startdash")
@click.option("-n", "--name", prompt="Dashboard name", help="Dashboard name.")
@click.option(
    "-p",
    "--dash-dir-path",
    prompt="Dashboards directory",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    default=Path(DEFAULT_DASHBOARDS_DIR),
    help="Path to the dashboards directory",
)
def start_dash(name: str, dash_dir_path: Path):
    name = valid_dir_name(name)
    dash_path = dash_dir_path / name

    validate_path_does_not_exist(dash_path)

    dash_path.mkdir()

    init_path = dash_path / "__init__.py"
    init_path.touch()

    dash_app_path = dash_path / "app.py"
    with dash_app_path.open("w") as fp:
        fp.writelines(BASE_APP_CONTENT.format(dash_name=name))



import shutil
from pathlib import Path

import click

from .config import DEFAULT_ASSETS_DIR

_ASSETS_SRC_DIR = (
    Path(__file__).parent.parent.resolve() / DEFAULT_ASSETS_DIR
).resolve()


def generate_assets(assets_dest_dir: Path):
    shutil.copytree(_ASSETS_SRC_DIR, assets_dest_dir, dirs_exist_ok=True)


@click.command(name="generate-assets")
@click.option(
    "-d",
    "--assets-dir",
    prompt="Assets directory",
    type=click.Path(file_okay=False, path_type=Path),
    default=Path(DEFAULT_ASSETS_DIR),
    help="Path to the dashboards directory",
)
def generate_assets_cmd(assets_dir: Path):
    if assets_dir.exists():
        new_assets = [
            f.relative_to(_ASSETS_SRC_DIR)
            for f in _ASSETS_SRC_DIR.rglob("*")
            if f.is_file()
        ]

        assets_overwrite = [
            f
            for f in assets_dir.rglob("*")
            if f.is_file() and f.relative_to(assets_dir) in new_assets
        ]

        file_list = "\n".join([f"  * {f}" for f in assets_overwrite])
        msg = (
            "This command will overwrite all changes you made in the "
            f"following files:\n{file_list}"
            "\nDo you want to continue?"
        )
        click.confirm(msg, abort=True)
    generate_assets(assets_dir)



import click

from ..app import DashboardApp


@click.command(name="runserver")
@click.option("-h", "--host", type=str, default="127.0.0.1")
@click.option("-p", "--port", type=int, default=8000)
@click.option("-d", "--debug", default=False, is_flag=True)
@click.pass_obj
def run_server(app: DashboardApp, host: str, port: int, debug: bool):
    app.run_server(host=host, port=port, debug=debug)



import setuptools

with open("README.md", "r") as fp:
    long_description = fp.read()


setuptools.setup(
    name="dash-dashboards",
    version="0.2.3",
    author="Filip Beć",
    description="Packakge to build Dash applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="TODO",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    include_package_data=True,
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    install_requires=[
        "click",
        "dash-bootstrap-components",
    ],
    extras_require={
        "tests": [
            "isort[colors]",
            "pytest",
            "pytest-cov",
            "pytest-env",
            "flake8",
            "black",
            "tox",
        ],
    },
    entry_points={
        "console_scripts": ["dash-admin = dash_dashboards.__main__:cli"],
    },
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Union

import dash_bootstrap_components as dbc
from dash import dcc, html
from dash.development.base_component import Component

ItemLayoutClass = Union[dcc.Link, dbc.Accordion]


class BaseMenuItem(ABC):
    @abstractmethod
    def get_menu_item_layout(self) -> ItemLayoutClass:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def pages(self) -> List["MenuItem"]:
        ...  # pragma: no cover


@dataclass
class MenuItem(BaseMenuItem):
    name: str
    layout: Union[Component, Callable[[None], Component]]
    route: str
    icon: str = None

    def get_menu_item_layout(self) -> ItemLayoutClass:
        return html.Li(
            className="nav-item",
            children=[
                dcc.Link(
                    href=self.route,
                    className="nav-link",
                    children=[html.I(className=self.icon), html.Span(self.name)],
                )
            ],
        )

    def get_layout(self) -> Component:
        return self.layout() if callable(self.layout) else self.layout

    @property
    def pages(self) -> List["MenuItem"]:
        return [self]


@dataclass
class MenuGroup(BaseMenuItem):
    name: str
    items: List[MenuItem]

    def get_menu_item_layout(self) -> ItemLayoutClass:
        return html.Li(
            className="nav-item",
            children=[
                html.H6(
                    className="sidebar-heading px-3 mt-4 mb-1 text-muted",
                    children=self.name,
                ),
                html.Ul(
                    className="nav flex-column mb-3 px-3",
                    children=[item.get_menu_item_layout() for item in self.items],
                ),
            ],
        )

    @property
    def pages(self) -> List["MenuItem"]:
        return self.items


def get_main_layout(navigation_layout: List[Component]) -> Component:
    return html.Div(
        className="container-fluid",
        children=[
            dcc.Location(id="url", refresh=False),
            html.Nav(
                className="d-flex flex-column col-4 col-md-3 col-lg-2 bg-white sidebar",
                children=navigation_layout,
            ),
            html.Div(
                className="ms-auto col-8 col-md-9 col-lg-10 px-md-4 py-3",
                id="content",
            ),
        ],
    )


Menu = List[BaseMenuItem]
